unit FiltersObjects;

interface

uses
  Windows, Classes, sysUtils, FiltApi, Graphics, contnrs;

type
  TFilterCallbackEvent = procedure( Sender : TObject; PercentDone : Integer ) of object;

  EFilterError = class( Exception );

  TAldus32Filter = class( TObject )
  private
    FDLLHandle : HINST;
    FInfoProc : GetFilerInfo;
    FRegisterProc : RegisterPercentCallback;
    FSetFilterPref : SetFilterPref;

    FPrefmem : THandle;

    FLibPath : String;
    FExtension : string;
    FLibName : string;

    FImportCallback : TFilterCallbackEvent;
    FShowProgressDialog: Boolean;
  protected
    procedure RaiseFilterError( code : integer );
    procedure DoCallback( Percent : integer );
    procedure SetFunctions; virtual;
    procedure ReleaseFunctions; virtual;
    procedure SetProperties; virtual;
  public
    property OnCallback : TFilterCallbackEvent read FImportCallback write FImportCallback;

    property Name : string read FLibName;
    property Path : string read FLibPath;
    property Extension : string read FExtension;

    property ShowProgressDialog : Boolean read FShowProgressDialog write FShowProgressDialog;

    constructor Create( const RegPath : string ); virtual;
  end;

  TImportFilter = class( TAldus32Filter )
  private
    FImportProc : ImportGr;

    FExtensions : TStringList;

    function GetExtension( index : integer ) : String;
    function GetExtensionCount: integer;
  protected
    procedure SetFunctions; override;
  public
    property Extension[ index : integer ] : string read GetExtension;
    property ExtensionCount : integer read GetExtensionCount;

    function ImportGraphic( const FileName : string ) : TMetafile;
    function IsCorrectExtension( const FileName : string ) : Boolean;

    constructor Create( const RegPath : string ); override;
    destructor Destroy; override;
  end;

  TImportGraphic = class( TObject )
  private
    FImportFilters : TObjectList;
    FExtensions : TStringList;
    FProgress: TFilterCallbackEvent;
    FProgressDialog: boolean;

    procedure InitializeFilters;
    procedure Callback( Sender : TObject; PercentDone : integer );
    function GetExtensionCount: integer;
    function GetSupportedExtension(index: integer): string;
  public
    property SupportedExtensions[ index : integer ] : string read GetSupportedExtension;
    property ExtensionCount : integer read GetExtensionCount;

    property ShowProgressDialog : boolean read FProgressDialog write FProgressDialog;

    function ImportGraphic( const FileName : string ) : TMetafile;

    constructor Create;
    destructor Destroy; override;
  published
    property OnImportProgress : TFilterCallbackEvent read FProgress write FProgress;
  end;

  TExportFilter = class( TAldus32Filter )
  private
    FExportProc : ExportGr;
    FQuality: integer;
    procedure SetQuality(const Value: integer);
  protected
    procedure SetFunctions; override;
    procedure SetProperties; override;
  public
    property Quality : integer read FQuality write SetQuality;

    procedure ExportGraphic( AGraphic : TMetafile; const FileName : String ); overload;
    procedure ExportGraphic( AGraphic : TGraphic; const FileName : string ); overload;

    function IsCorrectExtension( const FileName : string ) : Boolean;

    constructor Create( const RegPath : string ); override;
    destructor Destroy; override;
  end;

  TExportGraphic = class( TObject )
  private
    FExportFilters : TObjectList;
    FExtensions : TStringList;
    FQuality : integer;
    FProgress: TFilterCallbackEvent;
    FProgressDialog: boolean;

    procedure Callback( Sender : TObject; PercentDone : integer );
    procedure InitializeFilters;

    function GetExtensionCount: integer;
    function GetSupportedExtension(index: integer): string;
    procedure SetQuality(const Value: integer);
  public
    property SupportedExtensions[ index : integer ] : string read GetSupportedExtension;
    property ExtensionCount : integer read GetExtensionCount;

    property ShowProgressDialog : boolean read FProgressDialog write FProgressDialog;
    property Quality : integer read FQuality write SetQuality;

    procedure ExportGraphic( AGraphic : TGraphic; const FileName : string ); overload;
    procedure ExportGraphic( AGraphic : TMetafile; const FileName : string ); overload;

    constructor Create;
    destructor Destroy; override;
  published
    property OnImportProgress : TFilterCallbackEvent read FProgress write FProgress;
  end;


implementation

uses
  registry, syncobjs;

var
  CritSect : TCriticalSection;

const
  RegBaseImportPath = '\SOFTWARE\Microsoft\Shared Tools\Graphics Filters\Import';
  RegBaseExportPath = '\SOFTWARE\Microsoft\Shared Tools\Graphics Filters\Export';


function CallbackWrapper(lPct : Longint; pvData : Pointer ) : boolean; stdcall;
begin
  if pvData <> nil then begin
    TAldus32Filter( pvData^ ).DoCallback( lPct );
  end;

  Result := True;
end;

{ TImportFilter }

constructor TImportFilter.Create(const RegPath: string);
var
  Reg : TRegistry;
begin
  inherited;

  FExtensions := TStringList.Create;

  Reg := TRegistry.Create( KEY_READ );

  try
    // Read properties for import purposes
    with Reg do begin
      RootKey := HKEY_LOCAL_MACHINE;

      if OpenKeyReadOnly( RegPath ) = False then begin
        raise ERegistryException.CreateFmt( 'Failed to open "%s"!', [ RegPath ] );
      end;

      try
        FExtensions.CommaText := ReadString( 'ExtensionsEx' );
      except
        FExtensions.Text := '';
      end;
    end;
  finally
    Reg.CloseKey;
    Reg.Free;
  end;

  FDLLHandle := 0;
end;

destructor TImportFilter.Destroy;
begin
  ReleaseFunctions;

  if FExtensions <> nil then FExtensions.Free;

  inherited;
end;

function TImportFilter.GetExtension(index: integer): String;
begin
  Result := FExtensions[ index ];
end;

function TImportFilter.GetExtensionCount: integer;
begin
  Result := FExtensions.Count;
end;

function TImportFilter.ImportGraphic(const FileName: string): TMetafile;
var
  hdcPrint : HDC;
  FileInfo : FILESPEC;
  ImportInfo : GRPIX;
  ImportProcResult : integer;
begin
  SetFunctions;

  try
    CritSect.Enter;
    try
      SetProperties;
    finally
      CritSect.Leave;
    end;

    Result := TMetafile.Create;

    hdcPrint := GetDc( 0 );

    try
      with FileInfo do begin
        wFlags := 0;
        strLcopy( fxExtension.rgchExt, PChar( ExtractFileExt( FileName ) ) + 1,
                  sizeof( fxExtension ) - 1 ); // without point!!
        StrLCopy( strName, PChar( FileName ), sizeof( strName ) - 1 );
        dcbFile := 0;
        dcbFileHigh := 0;
      end;

      ZeroMemory( @ImportInfo, sizeof( GRPIX ) );
      ImportInfo.cbGrpix := sizeof( ImportInfo );

      CritSect.Enter;
      try
        ImportProcResult := FImportProc( hdcPrint, @FileInfo, @ImportInfo, FPrefmem );
      finally
        CritSect.Leave;
      end;

      if ImportProcResult = IE_NO_ERROR then begin
        if ImportInfo.grt = grtEMF then begin
          Result.Handle := ImportInfo.data.hemf;
        end else begin
          raise EConvertError.Create( 'The file has not been converted to EMF Format!' );
        end;
      end else begin
        RaiseFilterError( ImportProcResult );
      end;
    finally
      ReleaseDC( 0, hdcPrint );
    end;
  finally
    ReleaseFunctions;
  end;
end;

function TImportFilter.IsCorrectExtension(const FileName: string): Boolean;
var
  extension : String;
  i : Integer;
begin
  extension := LowerCase( Copy( ExtractFileExt( FileName ), 2, 255 ) );

  if FExtensions.Count = 0 then begin
    Result := extension = LowerCase( FExtension );
  end else begin
    Result := False;
    for i := 0 to FExtensions.Count - 1 do begin
      Result := Result or (extension = LowerCase( FExtensions[ i ] ) );
    end;
  end;
end;

procedure TImportFilter.SetFunctions;
var
  FilterInfo : Integer;
begin
  inherited;

  FPrefmem := 0;
  FilterInfo := FInfoProc( 3, nil, @FPrefmem, grfSupportEMF );

  if ((FilterInfo AND grfImport) <> grfImport) AND
     (FilterInfo <> 2) then begin
    raise EResNotFound.Create( 'Filter does not support Importing!' );
  end;

  @FImportProc := GetProcAddress( FDllHandle, 'ImportGr' );
  if @FImportProc = nil then begin
    raise EResNotFound.Create( 'Error extracting ImportProcedure' );
  end;
end;

{ TAldus32Filter }

constructor TAldus32Filter.Create(const RegPath: string);
var
  Reg : TRegistry;
begin
  FShowProgressDialog := True;

  Reg := TRegistry.Create( KEY_READ );

  try
    // Read common properties from both: import and export filters!
    with Reg do begin
      RootKey := HKEY_LOCAL_MACHINE;

      if OpenKeyReadOnly( RegPath ) = False then begin
        raise ERegistryException.CreateFmt( 'Failed to open "%s"!', [ RegPath ] );
      end;
      FLibName := ReadString( 'Name' );
      FLibPath := ReadString( 'Path' );
      FExtension := ReadString( 'Extensions' );
    end;
  finally
    Reg.CloseKey;
    Reg.Free;
  end;
end;

procedure TAldus32Filter.DoCallback(Percent: integer);
begin
  if Assigned( FImportCallback ) then FImportCallback( self, Percent );
end;

procedure TAldus32Filter.RaiseFilterError(code: integer);
begin
  case code of
    IE_INTERNAL_ERROR:  raise EFilterError.Create( 'Internal Error' );
    IE_NOT_MY_FILE:     raise EFilterError.Create( 'Invalid Graphic (not my file error)' );
    IE_TOO_BIG:         raise EFilterError.Create( 'Bitmap or Pict too big!' );
    IE_DUMB_BITMAP:     raise EFilterError.Create( 'Bitmap all white' );
    IE_BAD_VCHAR:       raise EFilterError.Create( 'Bad VCHAR!' );
    IE_BAD_TOKEN:       raise EFilterError.Create( 'Bad Token!' );
    IE_NO_VERIFY:       raise EFilterError.Create( 'Cannot Verify file' );
    IE_UNKNOWN_TYPE:    raise EFilterError.Create( 'Unknown file type!' );
    IE_NOT_WP_FILE:     raise EFilterError.Create( 'Not a wp file' );
    IE_BAD_FILE_DATA:   raise EFilterError.Create( 'Bad File Data' );
    IE_IMPORT_ABORT:    raise EFilterError.Create( 'Import Aborted' );
    IE_MEM_FULL:        raise EFilterError.Create( 'Run out of memory during import!' );
    IE_MSNG_FONTS:      raise EFilterError.Create( 'System font not found!' );
    IE_META_TOO_BIG:    raise EFilterError.Create( 'Metafile too big!' );
    IE_MEM_FAIL:        raise EFilterError.Create( 'Couldnt lock memory during import' );
    IE_NO_FILTER:       raise EFilterError.Create( 'Expected Filter not found' );
    IE_UNSUPP_COMPR:    raise EFilterError.Create( 'Unsupported Compression style' );
    IE_UNSUPP_VERSION:  raise EFilterError.Create( 'Unsupported Version' );
    IE_UNSUPP_COLOR:    raise EFilterError.Create( 'Unsupported Color style' );

    IE_ERROR_NOMSG:     raise EFilterError.Create( 'Dialog Box Cancel' );
    IE_FILE_NOT_FOUND:  raise EFilterError.Create( 'File not found' );
    IE_BUG:             raise EFilterError.Create( 'Bug!' );
    IE_BAD_METAFILE:    raise EFilterError.Create( 'Inconsistent metafile data' );
    IE_BAD_METAFILE2:   raise EFilterError.Create( 'Inconsistent metafile data' );

    IE_BAD_PARAM:       raise EFilterError.Create( 'Error bad Parameter / Parameter is not supported!' );
    IE_UNSUPP_FORMAT:   raise EFilterError.Create( 'Cannot provide/accept format' );
    FA_DISK_ERROR:      raise EFilterError.Create( 'Disk Error!' );
  else
    raise EFilterError.CreateFmt( 'Unknown Error: %d', [ Code ] );
  end;
end;

procedure TAldus32Filter.ReleaseFunctions;
begin
  if FDLLHandle <> 0 then begin
    FreeLibrary( FDLLHandle );
    FDLLHandle := 0;
  end;
end;

procedure TAldus32Filter.SetFunctions;
begin
  FDllHandle := LoadLibrary( PChar( FLibPath ) );
  if FDLLHandle < 32 then begin
    raise EResNotFound.CreateFmt( 'Error Loading Library "%s"', [ FLibPath ] );
  end;

  @FInfoProc := GetProcAddress( FDllHandle, 'GetFilterInfo' );
  if @FInfoProc = nil then begin
    raise EResNotFound.Create( 'Error extracting InfoProcedure' );
  end;

  @FRegisterProc := GetProcAddress( FDllHandle, 'RegisterPercentCallback' );
  if @FRegisterProc <> nil then begin
    // try to register a procedure!!, just try
    FRegisterProc( FPrefmem, @CallbackWrapper, @self );
  end;

  @FSetFilterPref := GetProcAddress( FDllHandle, 'SetFilterPref' );
end;

procedure TAldus32Filter.SetProperties;
var
  setFilterResult : integer;
  test : integer;
begin
  if @FSetFilterPref <> nil then begin
    if FShowProgressDialog then begin
      setFilterResult := FSetFilterPref( FPrefmem, FiltOptProgressDialog, PChar( strYes ),
                                         Length( strYes ) + 1, REG_SZ );
    end else begin
      setFilterResult := FSetFilterPref( FPrefmem, FiltOptProgressDialog, PChar( strNo ),
                                         Length( strNo ) + 1, REG_SZ );
    end;


    // don't raise an error when this parameter is not supported!!!
    if (setFilterResult <> IE_NO_ERROR) AND (setFilterResult <> IE_BAD_PARAM) then begin
      RaiseFilterError( setFilterResult );
    end;
  end;
end;

{ TImportGraphic }

procedure TImportGraphic.Callback(Sender: TObject; PercentDone: integer);
begin
  if Assigned( FProgress ) then FProgress( self, PercentDone );
end;

constructor TImportGraphic.Create;
begin
  inherited;

  FExtensions := TStringList.Create;
  FImportFilters := TObjectList.Create( True );

  FProgressDialog := True;

  InitializeFilters;
end;

destructor TImportGraphic.Destroy;
begin
  inherited;

  if FExtensions <> nil then FExtensions.Free;
  if FImportFilters <> nil then FImportFilters.Free;
end;

function TImportGraphic.GetExtensionCount: integer;
begin
  Result := FExtensions.Count;
end;

function TImportGraphic.GetSupportedExtension(index: integer): string;
begin
  Result := FExtensions[ index ];
end;

function TImportGraphic.ImportGraphic(const FileName: string): TMetafile;
var
  i : integer;
  AObject : TImportFilter;
begin
  for i := 0 to FImportFilters.Count - 1 do begin
    AObject := FImportFilters[ i ] as TImportFilter;

    if AObject.IsCorrectExtension( FileName ) then begin
      AObject.ShowProgressDialog := FProgressDialog;
      Result := AObject.ImportGraphic( FileName );
      exit;
    end;
  end;

  raise EFilterError.CreateFmt( 'Filetype is not supported: "%s"', [ FileName ] );
end;

procedure TImportGraphic.InitializeFilters;
var
  Reg : TRegistry;
  RegEntries : TStringList;
  i : integer;
  ImportObject : TImportFilter;
begin
  Reg := TRegistry.Create( KEY_READ );
  try
    with Reg do begin
      RootKey := HKEY_LOCAL_MACHINE;
      if OpenKeyReadOnly( RegBaseImportPath ) = False then begin
        raise ERegistryException.CreateFmt( 'Failed to open "%s"!', [ RegBaseImportPath ] );
      end;

      RegEntries := TStringList.Create;
      try
        GetKeyNames( RegEntries );

        for i := 0 to RegEntries.Count - 1 do begin
          ImportObject := TImportFilter.Create( RegBaseImportPath + '\' + RegEntries[ i ] );
          ImportObject.OnCallback := Callback;
          FImportFilters.Add( ImportObject );

          // ShortCut ;))
          FExtensions.AddStrings( ImportObject.FExtensions );
        end;
      finally
        RegEntries.Free;
      end;
    end;
  finally
    Reg.CloseKey;
    Reg.Free;
  end;
end;

{ TExportFilter }

constructor TExportFilter.Create(const RegPath: string);
begin
  inherited;

  FQuality := 75; // 75 percent!!
end;

destructor TExportFilter.Destroy;
begin
  inherited;

  ReleaseFunctions;
end;

procedure TExportFilter.ExportGraphic(AGraphic: TGraphic;
  const FileName: string);
var
  Metafile : TMetafile;
  MetaCanvas : TMetafileCanvas;
begin
  if AGraphic = nil then raise EFilterError.Create( 'AGraphic may not be nil!' );

  Metafile := TMetafile.Create;
  try
    Metafile.Width := AGraphic.Width;
    Metafile.Height := AGraphic.Height;
    MetaCanvas := TMetafileCanvas.Create( Metafile, 0 ); //ScreenDC
    try
      MetaCanvas.Draw( 0, 0, AGraphic );
    finally
      MetaCanvas.Free;
    end;

    ExportGraphic( MetaFile, FileName );
  finally
    Metafile.Free;
  end;
end;

procedure TExportFilter.ExportGraphic(AGraphic: TMetafile;
  const FileName: String);
var
  fileInfo : TFileSpec;
  metaInfo : TGRPIX;
  ExportInfo : integer;
begin
  if AGraphic = nil then raise EFilterError.Create( 'Graphic may not be nil!' );

  SetFunctions;

  try
    CritSect.Enter;
    try
      SetProperties;
    finally
      CritSect.Leave;
    end;

    ZeroMemory( @fileInfo, sizeof( fileInfo ) );
    ZeroMemory( @metaInfo, sizeof( metaInfo ) );
    with fileInfo do begin
      StrCopy( strName, PChar( FileName + '.' + FExtension ) );
      StrCopy( fxExtension.rgchExt, PChar( FExtension ) );
    end;

    with metaInfo do begin
      cbGrpix := sizeof( metaInfo );
      grt := grtEMF;
      fPointer := 0;  // false
      data.hmf := AGraphic.Handle;
      bbox := Rect( 0, 0, AGraphic.Width, AGraphic.Height );
      inch := AGraphic.Inch;
    end;

    CritSect.Enter;
    try
      ExportInfo := FExportProc( @fileInfo, @metaInfo, FPrefmem );
    finally
      CritSect.Leave;
    end;

    if ExportInfo <> IE_NO_ERROR then begin
      RaiseFilterError( ExportInfo );
    end;
  finally
    ReleaseFunctions;
  end;
end;

function TExportFilter.IsCorrectExtension(const FileName: string): Boolean;
var
  extension : String;
begin
  extension := LowerCase( Copy( ExtractFileExt( FileName ), 2, 255 ) );

  Result := extension = LowerCase( FExtension );
end;

procedure TExportFilter.SetFunctions;
var
  FilterInfo : integer;
begin
  inherited;

  FPrefmem := 0;
  FilterInfo := FInfoProc( 3, nil, @FPrefmem, grfSupportEMF );

  if ((FilterInfo AND grfExport) <> grfExport) AND
     (FilterInfo <> 2) then begin
    raise EResNotFound.Create( 'Filter does not support Exporting!' );
  end;

  @FExportProc := GetProcAddress( FDllHandle, 'ExportGr' );
  if @FExportProc = nil then begin
    raise EResNotFound.Create( 'Error extracting ExportProcedure' );
  end;
end;

procedure TExportFilter.SetProperties;
var
  setPropResult : integer;
begin
  inherited;

  if @FSetFilterPref <> nil then begin
    setPropResult := FSetFilterPref( FPrefmem, FiltOptQuality,
                     Pointer( @FQuality ), sizeof( FQuality ), REG_DWORD );

    if (setPropResult <> IE_NO_ERROR) AND (setPropResult <> IE_BAD_PARAM) then begin
      RaiseFilterError( setPropResult );
    end;
  end;
end;

procedure TExportFilter.SetQuality(const Value: integer);
begin
  FQuality := Value;

  if FQuality < 0 then begin
    FQuality := 0;
  end;

  if FQuality > 100 then begin
    FQuality := 100;
  end;
end;

{ TExportGraphic }

procedure TExportGraphic.Callback(Sender: TObject; PercentDone: integer);
begin
  if Assigned( FProgress ) then FProgress( self, PercentDone );
end;

constructor TExportGraphic.Create;
begin
  inherited;

  FExtensions := TStringList.Create;
  FExportFilters := TObjectList.Create( True );

  FProgressDialog := True;

  InitializeFilters;
end;

destructor TExportGraphic.Destroy;
begin
  inherited;

  if FExtensions <> nil then FExtensions.Free;
  if FExportFilters <> nil then FExportFilters.Free;
end;

procedure TExportGraphic.ExportGraphic(AGraphic: TGraphic;
  const FileName: string);
var
  i : integer;
  AObject : TExportFilter;
begin
  for i := 0 to FExportFilters.Count - 1 do begin
    AObject := FExportFilters[ i ] as TExportFilter;

    if AObject.IsCorrectExtension( FileName ) then begin
      AObject.ShowProgressDialog := FProgressDialog;
      AObject.Quality := FQuality;
      AObject.ExportGraphic( AGraphic, FileName );
      exit;
    end;
  end;

  raise EFilterError.CreateFmt( 'Filetype is not supported: "%s"', [ FileName ] );
end;

procedure TExportGraphic.ExportGraphic(AGraphic: TMetafile;
  const FileName: string);
var
  i : integer;
  AObject : TExportFilter;
begin
  for i := 0 to FExportFilters.Count - 1 do begin
    AObject := FExportFilters[ i ] as TExportFilter;

    if AObject.IsCorrectExtension( FileName ) then begin
      AObject.ShowProgressDialog := FProgressDialog;
      AObject.Quality := FQuality;
      AObject.ExportGraphic( AGraphic, FileName );
      exit;
    end;
  end;

  raise EFilterError.CreateFmt( 'Filetype is not supported: "%s"', [ FileName ] );
end;

function TExportGraphic.GetExtensionCount: integer;
begin
  Result := FExtensions.Count;
end;

function TExportGraphic.GetSupportedExtension(index: integer): string;
begin
  Result := FExtensions[ index ];
end;

procedure TExportGraphic.InitializeFilters;
var
  Reg : TRegistry;
  RegEntries : TStringList;
  i : integer;
  ExportObject : TExportFilter;
begin
  Reg := TRegistry.Create( KEY_READ );
  try
    with Reg do begin
      RootKey := HKEY_LOCAL_MACHINE;
      if OpenKeyReadOnly( RegBaseExportPath ) = False then begin
        raise ERegistryException.CreateFmt( 'Failed to open "%s"!', [ RegBaseExportPath ] );
      end;

      RegEntries := TStringList.Create;
      try
        GetKeyNames( RegEntries );

        for i := 0 to RegEntries.Count - 1 do begin
          ExportObject := TExportFilter.Create( RegBaseExportPath + '\' + RegEntries[ i ] );
          ExportObject.OnCallback := Callback;
          FExportFilters.Add( ExportObject );

          FExtensions.Add( ExportObject.Extension );
        end;
      finally
        RegEntries.Free;
      end;
    end;
  finally
    Reg.CloseKey;
    Reg.Free;
  end;
end;

procedure TExportGraphic.SetQuality(const Value: integer);
begin
  FQuality := Value;

  if FQuality < 0 then begin
    FQuality := 0;
  end;

  if FQuality > 100 then begin
    FQuality := 100;
  end;
end;

initialization
  try
    CritSect := TCriticalSection.Create;
  except
    CritSect := nil;
    exit;
  end;

finalization
  if Assigned( CritSect ) then CritSect.Free;

end.
